#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#------------------------------------------------------------------------------------------------
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# -----------------------------------------------------------------------------------------------
# THIS PLUGIN WAS DEVELOPED AND TESTED UNDER WINDOWS VISTA ONLY AND MAY OR MAY NOT WORK ON A MAC. 
# -----------------------------------------------------------------------------------------------
# THIS PLUGIN IS ANOTHER FIGMENT OF MY IMAGENATION AND IS NOT BASED ON OR COMPLY WITH ANYTHING! 
# -----------------------------------------------------------------------------------------------
#    Name:	Followme_Multiple
#      By:	sdmitch
#   Usage:	To extrude a single profile around multiple pre-selected faces and/or edges
#		 Note:	The profile MUST be the smallest face selected and created at the model origin.  The
#						profile placed on the edge relative to its position to the model origin.
#    Date:	Jan 2013
#------------------------------------------------------------------------------------------------
#
require 'Sketchup'
#
module SDM

	module Followme_Multiple

		if not file_loaded?(File.basename(__FILE__))
			UI.menu('Plugins').add_item('Followme_Multiple') { self.main }
			file_loaded(File.basename(__FILE__))
		end

		def self.main
			mod = Sketchup.active_model
			ent = mod.active_entities
			sel = mod.selection
			mod.start_operation "Followme Mulitple"
			#
			faces=[]; paths=[]; org=Geom::Point3d.new()
			begin
				if sel.first.class==Sketchup::Face
					faces<<sel.first;
					sel.remove faces[-1].edges; sel.remove faces[-1]
				elsif sel.first.class==Sketchup::Edge && sel.first.curve
					paths<<sel.first.curve.edges
					sel.remove sel.first.curve.edges
				elsif sel.first.class==Sketchup::Edge && sel.first.faces.length==0
					paths<<sel.first.all_connected
					sel.remove sel.first.all_connected
				else
					sel.remove sel.first # not a valid entity so get rid of it
				end
			end until sel.empty?
			hsh={};faces.each{|f| hsh[f]=f.area}#collect face areas
			prof=hsh.sort{|a,b| a[1]<=>b[1]}[0][0]#profile is smallest
			# pp=prof.outer_loop.vertices.collect{|v| v.position}
			# hsh={};pp.each{|p| hsh[p]=p.distance(ORIGIN)}
			# cp=hsh.sort{|a,b| a[1]<=>b[1]}[0][0];cpi=pp.index(cp)
			#
			# Process all faces selected
			#
			for f in faces
				next if f==prof#skip if profile face
				f.reverse! if f.normal.z < 0; rev_pf=false
				edg=f.outer_loop.edges[0] # get starting edge
				sp,evec=edg.line;evec.length=edg.length/2
				tr=Geom::Transformation.new(sp.offset(evec))
				pp=prof.outer_loop.vertices.collect{|v| v.position}
				pp.each{|p| p.transform! tr}; cp=org.transform tr#move profile to start point
				px,py,pz=Geom.fit_plane_to_points pp
				pvec=Geom::Vector3d.new(px,py,pz)
				pvec.reverse! if evec.angle_between(pvec) > 90.degrees
				evec.z=0.0;pvec.z=0.0;
				norm=evec*pvec; #puts "\nperp norm = #{norm}"
				unless norm.length==0#check perpendicularity
					ang=evec.angle_between pvec; #puts "perpendicular check ang = #{ang.radians}"
					if ang > 5.degrees && ang < 175.degrees
						ang = ang - 180.degrees if ang > 90.degrees
						tr=Geom::Transformation.rotation(cp,norm,-ang)
						pp.each{|p| p.transform! tr}
					end
				end
				px,py,pz=Geom.fit_plane_to_points pp
				pvec=Geom::Vector3d.new(px,py,pz); norm=evec*pvec
				unless norm.length == 0.0 # check tilt
					ang=evec.angle_between pvec; #puts "tilt check ang = #{ang.radians}"
					if ang > 5.degrees && ang < 175.degrees
						ang = ang - 180.degrees if ang > 90.degrees
						tr=Geom::Transformation.rotation(cp,norm,-ang)
						pp.each{|p| p.transform! tr}
					end
				end
				if f.outer_loop.edges[0].reversed_in? f
					norm=evec.axes[1]; #puts "reversed edge norm = #{norm}"
					tr=Geom::Transformation.rotation(cp,norm,180.degrees)
					pp.each{|p| p.transform! tr};rev_pf=true
				end
				pf=ent.add_face(pp); pf.reverse! if rev_pf; #add profile face
				edges=f.outer_loop.edges#get edges of f
				pf.followme edges#do extrude
			end
			#
			# Process all selected Paths
			#
			for path in paths
				unless path[0].curve
					path=self.sort_edges(path)
					path.reverse! if path[0].start.edges.length>1
				end
				sp,evec=path[0].line
				tr=Geom::Transformation.new(sp)
				pp=prof.outer_loop.vertices.collect{|v| v.position}
				pp.each{|p| p.transform! tr};cp=org.transform tr
				px,py,pz=Geom.fit_plane_to_points pp
				pvec=Geom::Vector3d.new(px,py,pz)
				pvec.reverse! if evec.angle_between(pvec) > 90.degrees
				evec.z=0.0;pvec.z=0.0;norm=evec*pvec
				unless norm.length==0#check perpendicularity
					ang=evec.angle_between pvec; #puts "perpendicular check ang = #{ang.radians}"
					if ang > 5.degrees && ang < 175.degrees
						ang = ang - 180.degrees if ang > 90.degrees
						tr=Geom::Transformation.rotation(cp,norm,-ang)
						pp.each{|p| p.transform! tr}
					end
				end
				px,py,pz=Geom.fit_plane_to_points pp
				pvec=Geom::Vector3d.new(px,py,pz); norm=evec*pvec
				unless norm.length == 0.0 # check tilt
					ang=evec.angle_between pvec; #puts "tilt check ang = #{ang.radians}"
					if ang > 5.degrees && ang < 175.degrees
						ang = ang - 180.degrees if ang > 90.degrees
						tr=Geom::Transformation.rotation(cp,norm,-ang)
						pp.each{|p| p.transform! tr}
					end
				end
				pf=ent.add_face(pp); pf.reverse!
				pf.followme path
			end
			#
			mod.commit_operation
		end#def main
		
		def self.sort_edges(edges)
			re=[]
			while edges.length>0
				start=nil
				edges.each{|x|
				start=x.start; break  if (x.start.edges&edges).length==1
				start=x.end; break if (x.end.edges&edges).length==1
				}
				curve=[]
				while start
					next_edge=(start.edges&edges-curve)[0]
					if next_edge
						curve << next_edge
						start=next_edge.other_vertex(start)
					else
						start=nil
					end
				end
				edges-=curve
				re += curve 
			end
			return re
		end
		
	end#module Followme_Multiple
	
end#module SDM
